import dsp
import random

from dsp.utils import dotdict, has_answer, DPR_normalize, normalize_text, EM


class Example(dotdict):
    def __init__(self, *args, **kwargs):
        assert len(args) <= 1
        super().__init__()

        if len(args):
            assert len(args) == 1
            self.update(args[0])
        
        self.update(**kwargs)
    
    def copy(self, **kwargs):
        the_copy = Example(**{**dict(self), **kwargs})

        return the_copy

    def without(self, *keys):
        keys = set(keys)
        return Example({k: v for k, v in self.items() if k not in keys})

    def demos_at(self, fn):
        def at(d):
            try:
                return fn(d).without('augmented')
            except:
                return {}

        demos = [d.copy(**at(d)) for d in self.demos]
        return self.copy(demos=demos)

    # def __str__(self):
    #     return self.get('question', self.get('query'))





def annotate(*transformations):
    def do_augment(train, k=None, return_all=False):
        rdemos = []
        ademos = []

        for example in (train): # tqdm.tqdm
            raw_example = dsp.Example(example)

            if k and len(ademos) >= k:
                example = None

            for f in transformations:
                if example is None:
                    break
                
                example = f(example)

            if example is not None:
                example.augmented = True
                ademos.append(example)
            else:
                raw_example.augmented = False
                rdemos.append(raw_example)
        
        if return_all:
            return ademos + rdemos

        return ademos
    
    return do_augment

def sample(train, k: int):
    branch_idx = hasattr(dsp.settings, 'branch_idx') and isinstance(dsp.settings.branch_idx, int) and dsp.settings.branch_idx
    branch_idx = branch_idx or 0
    # print(f"branch_idx = {branch_idx}")

    rng = random.Random(branch_idx)
    shuffled_train = list(train)
    rng.shuffle(shuffled_train)

    subset = shuffled_train[:k]
    subset = [dsp.Example(x) for x in subset] # augmented=False

    return subset

def all_but(train, x):
    # output = [y for y in train if y.question != x.question]

    output = [y for y in train
              if not set.intersection(set(x.get('history', []) + [x.question]), set(y.get('history', []) + [y.question]))]
    # print(len(output))

    return output


def passage_match(passages, answers):
    # passages = example.passages
    # answers = example.answers
    return any(passage_has_answers(psg, answers) for psg in passages)

def answer_match(prediction, answers):
    # pred = example.prediction
    # answers = example.answers
    return EM(prediction, answers)

def passage_has_answers(passage, answers):
    return has_answer([DPR_normalize(normalize_text(ans)) for ans in answers], normalize_text(passage))



import os
import tqdm
import datetime
import itertools

from collections import defaultdict


def print_message(*s, condition=True, pad=False, sep=None):
    s = ' '.join([str(x) for x in s])
    msg = "[{}] {}".format(datetime.datetime.now().strftime("%b %d, %H:%M:%S"), s)

    if condition:
        msg = msg if not pad else f'\n{msg}\n'
        print(msg, flush=True, sep=sep)


    return msg


def timestamp(daydir=False):
    format_str = f"%Y-%m{'/' if daydir else '-'}%d{'/' if daydir else '_'}%H.%M.%S"
    result = datetime.datetime.now().strftime(format_str)
    return result


def file_tqdm(file):
    print(f"#> Reading {file.name}")

    with tqdm.tqdm(total=os.path.getsize(file.name) / 1024.0 / 1024.0, unit="MiB") as pbar:
        for line in file:
            yield line
            pbar.update(len(line) / 1024.0 / 1024.0)

        pbar.close()


def create_directory(path):
    if os.path.exists(path):
        print('\n')
        print_message("#> Note: Output directory", path, 'already exists\n\n')
    else:
        print('\n')
        print_message("#> Creating directory", path, '\n\n')
        os.makedirs(path)


def deduplicate(seq):
    """
    Source: https://stackoverflow.com/a/480227/1493011
    """

    seen = set()
    return [x for x in seq if not (x in seen or seen.add(x))]


def batch(group, bsize, provide_offset=False):
    offset = 0
    while offset < len(group):
        L = group[offset: offset + bsize]
        yield ((offset, L) if provide_offset else L)
        offset += len(L)
    return


class dotdict(dict):
    """
    dot.notation access to dictionary attributes
    Credit: derek73 @ https://stackoverflow.com/questions/2352181
    """
    __getattr__ = dict.__getitem__
    __setattr__ = dict.__setitem__
    __delattr__ = dict.__delitem__


class dotdict_lax(dict):
    __getattr__ = dict.get
    __setattr__ = dict.__setitem__
    __delattr__ = dict.__delitem__


def flatten(L):
    # return [x for y in L for x in y]

    result = []
    for _list in L:
        result += _list

    return result


def zipstar(L, lazy=False):
    """
    A much faster A, B, C = zip(*[(a, b, c), (a, b, c), ...])
    May return lists or tuples.
    """

    if len(L) == 0:
        return L

    width = len(L[0])

    if width < 100:
        return [[elem[idx] for elem in L] for idx in range(width)]

    L = zip(*L)

    return L if lazy else list(L)


def zip_first(L1, L2):
    length = len(L1) if type(L1) in [tuple, list] else None

    L3 = list(zip(L1, L2))

    assert length in [None, len(L3)], "zip_first() failure: length differs!"

    return L3


def int_or_float(val):
    if '.' in val:
        return float(val)
        
    return int(val)


def groupby_first_item(lst):
    groups = defaultdict(list)

    for first, *rest in lst:
        rest = rest[0] if len(rest) == 1 else rest
        groups[first].append(rest)

    return groups


def process_grouped_by_first_item(lst):
    """
        Requires items in list to already be grouped by first item.
    """

    groups = defaultdict(list)

    started = False
    last_group = None

    for first, *rest in lst:
        rest = rest[0] if len(rest) == 1 else rest

        if started and first != last_group:
            yield (last_group, groups[last_group])
            assert first not in groups, f"{first} seen earlier --- violates precondition."

        groups[first].append(rest)

        last_group = first
        started = True

    return groups


def grouper(iterable, n, fillvalue=None):
    """
    Collect data into fixed-length chunks or blocks
        Example: grouper('ABCDEFG', 3, 'x') --> ABC DEF Gxx"
        Source: https://docs.python.org/3/library/itertools.html#itertools-recipes
    """

    args = [iter(iterable)] * n
    return itertools.zip_longest(*args, fillvalue=fillvalue)


def lengths2offsets(lengths):
    offset = 0

    for length in lengths:
        yield (offset, offset + length)
        offset += length

    return


# see https://stackoverflow.com/a/45187287
class NullContextManager(object):
    def __init__(self, dummy_resource=None):
        self.dummy_resource = dummy_resource
    def __enter__(self):
        return self.dummy_resource
    def __exit__(self, *args):
        pass


def load_batch_backgrounds(args, qids):
    if args.qid2backgrounds is None:
        return None

    qbackgrounds = []

    for qid in qids:
        back = args.qid2backgrounds[qid]

        if len(back) and type(back[0]) == int:
            x = [args.collection[pid] for pid in back]
        else:
            x = [args.collectionX.get(pid, '') for pid in back]

        x = ' [SEP] '.join(x)
        qbackgrounds.append(x)
    
    return qbackgrounds



from dsp.templates import TemplateV2, passages2text, format_answers, Field

class Type:
    def __init__(self, prefix, desc, format=None, aliases=[]) -> None:
        self.prefix = prefix
        self.desc = desc
        self.format = format
        self.aliases = aliases

        assert aliases == [], "TODO: If aliases is used, handling of input_variable in V2 needs to be richer."
    
    def __call__(self, **kwargs):
        kwargs = {**self.__dict__, **kwargs}
        return Type(**kwargs)

class Template(TemplateV2):
    def __init__(self, instructions, **kwargs):
        self.instructions = instructions

        self.fields = []
        self.format_handlers = {"context": passages2text, "answers": format_answers}

        for key, value in kwargs.items():
            prefix = value.prefix
            separator = ' ' if prefix.rstrip() == prefix else prefix[len(prefix.rstrip()):]
            field = Field(name=prefix.strip(), description=value.desc,
                          input_variable=key, output_variable=key, seperator=separator)
            self.fields.append(field)

            if value.format:
                for name in [key, *value.aliases]:
                    self.format_handlers[name] = value.format



from collections import namedtuple
import re

import dsp

Field = namedtuple("Field", "name seperator input_variable output_variable description")

# TODO: de-duplicate with dsp/templates/template.py


def passages2text(passages):
    if type(passages) is str:
        return passages

    assert type(passages) in [list, tuple]

    if len(passages) == 1:
        return f"«{passages[0]}»"

    return "\n".join([f"[{idx+1}] «{txt}»" for idx, txt in enumerate(passages)])


def format_answers(answers):
    if isinstance(answers, list):
        if len(answers) >= 1:
            return str(answers[0]).strip()
        elif len(answers) == 0:
            raise ValueError(f"No answers found")
    elif isinstance(answers, str):
        return answers
    else:
        raise ValueError(f"Unable to parse answers of type {type(answers)}")


class TemplateV2:
    def __init__(self, template, format_handlers={"passages": passages2text, "context": passages2text, "answer": format_answers, "answers": format_answers}):
        self.format_handlers = format_handlers

        template = template.strip()

        self.instructions = re.search("(.*)\n", template).group(1)
        template = template[len(self.instructions):].strip()

        self.fields = []
        while len(template) > 0:
            match = re.search("(.*)(\s){(.*)}\s(.*\${.*})", template)
            if match is not None:
                name = match.group(1)
                seperator = match.group(2)
                variable = match.group(3)
                description = match.group(4)
            else:
                match = re.search("(.*)(\s){(.*)}", template)
                if match is not None:
                    name = match.group(1)
                    seperator = match.group(2)
                    variable = match.group(3)
                    description = None
                else:
                    raise ValueError(f"Could not parse template")

            var_match = re.match("(.*) -> (.*)", variable)
            if var_match is not None:
                input_variable = var_match.group(1)
                output_variable = var_match.group(2)
            else:
                input_variable = variable
                output_variable = variable

            self.fields.append(
                Field(
                    name=name,
                    seperator=seperator,
                    input_variable=input_variable,
                    output_variable=output_variable,
                    description=description,
                )
            )

            template = template[len(match.group(0)):].strip()

    def query(self, example, is_demo=False):
        result = []

        if not is_demo:
            has_value = [
                field.input_variable in example
                and example[field.input_variable] is not None
                and example[field.input_variable] != ""
                for field in self.fields
            ]

            for i in range(1, len(has_value)):
                if has_value[i - 1] and not any(has_value[i:]):
                    example[self.fields[i].input_variable] = ""
                    break

        for field in self.fields:
            if (field.input_variable in example and example[field.input_variable] is not None):
                if field.input_variable in self.format_handlers:
                    format_handler = self.format_handlers[field.input_variable]
                else:
                    def format_handler(x): return ' '.join(x.split())
                
                result.append(f"{field.name}{field.seperator}{format_handler(example[field.input_variable])}")

        if self._has_augmented_guidelines() and ("augmented" in example and example.augmented):
            return "\n\n".join(result)
        else:
            return "\n".join(result)

    def guidelines(self):
        result = "Follow the following format.\n\n"

        example = dsp.Example()
        for field in self.fields:
            example[field.input_variable] = field.description
        example.augmented = self._has_augmented_guidelines()

        result += self.query(example)
        return result

    def _has_augmented_guidelines(self):
        return len(self.fields) > 3 or any(field.seperator == "\n" for field in self.fields)

    def extract(self, example, raw_pred):
        example = dsp.Example(example)

        raw_pred = raw_pred.strip()

        idx = 0
        while idx < len(self.fields):
            if (self.fields[idx].input_variable not in example or example[self.fields[idx].input_variable] is None):
                break
            idx += 1
        
        idx = min(idx, len(self.fields) - 1)

        while raw_pred != "" and idx < len(self.fields):
            # print("HERE,", idx)
            if idx < len(self.fields) - 1:
                next_field_name = '\n' + self.fields[idx + 1].name
                offset = raw_pred.find(next_field_name)
                # print(f"Looking for ```{next_field_name}```. Found it at {offset}")

                if offset >= 0:
                    example[self.fields[idx].output_variable] = raw_pred[:offset].strip()
                    raw_pred = raw_pred[offset + len(next_field_name):].strip()
                    idx += 1
                else:
                    example[self.fields[idx].output_variable] = raw_pred.strip()
                    raw_pred = ""
                    idx += 1
                    # print("Exiting")
                    break

            else: #if idx == len(self.fields) - 1:
                assert idx == len(self.fields) - 1, (idx, len(self.fields))
                example[self.fields[idx].output_variable] = raw_pred.strip()
                break
                # offset = raw_pred.find(self.fields[idx].name)
                # if offset < 0:
                # else:
                #     raw_pred = raw_pred[offset + len(self.fields[idx].name):].strip()
                #     example[self.fields[idx].output_variable] = raw_pred

        return example

    def __call__(self, example):
        example = dsp.Example(example)

        if self.fields[-1].input_variable in example:
            del example[self.fields[-1].input_variable]

        rdemos = [
            self.query(demo, is_demo=True)
            for demo in example.demos
            if (
                ("augmented" not in demo or not demo.augmented)
                and (
                    self.fields[-1].input_variable in demo
                    and demo[self.fields[-1].input_variable] is not None
                )
            )
        ]
        ademos = [
            self.query(demo, is_demo=True)
            for demo in example.demos
            if "augmented" in demo and demo.augmented
        ]

        long_query = self._has_augmented_guidelines()
        if long_query:
            example["augmented"] = True
        query = self.query(example)

        rdemos = "\n\n".join(rdemos)
        if len(rdemos) >= 1 and len(ademos) == 0 and not long_query:
            rdemos_and_query = "\n\n".join([rdemos, query])
            parts = [self.instructions, self.guidelines(), rdemos_and_query]
        elif len(rdemos) == 0:
            parts = [self.instructions, self.guidelines(), *ademos, query]
        else:
            parts = [self.instructions, rdemos,
                     self.guidelines(), *ademos, query]

        prompt = "\n\n---\n\n".join([p.strip() for p in parts if p])

        return prompt.strip()



import openai
import openai.error
import backoff
import functools

from dsp.modules.cache_utils import CacheMemory, NotebookCacheMemory, cache_turn_on


def backoff_hdlr(details):
    # Handler from https://pypi.org/project/backoff/
    print("Backing off {wait:0.1f} seconds after {tries} tries "
          "calling function {target} with args {args} and kwargs "
          "{kwargs}".format(**details))


class GPT3:
    def __init__(self, model='text-davinci-002', api_key=None):
        if api_key:
            openai.api_key = api_key

        self.kwargs = {'model': model, 'temperature': 0.0, 'max_tokens': 150, 'top_p': 1,
                       'frequency_penalty': 0, 'presence_penalty': 0, 'n': 1}

        self.history = []

    def basic_request(self, prompt, **kwargs):
        raw_kwargs = kwargs
        kwargs = {**self.kwargs, 'prompt': prompt, **kwargs}
        response = cached_gpt3_request(**kwargs)

        history = {'prompt': prompt, 'response': response,
                   'kwargs': kwargs, 'raw_kwargs': raw_kwargs}
        self.history.append(history)

        return response

    @backoff.on_exception(backoff.expo,
                          (openai.error.RateLimitError,
                           openai.error.ServiceUnavailableError),
                          max_time=1000,
                          on_backoff=backoff_hdlr)
    def request(self, prompt, **kwargs):
        return self.basic_request(prompt, **kwargs)

    def print_green(self, text, end='\n'):
        print("\x1b[32m" + str(text) + "\x1b[0m", end=end)
    
    def print_red(self, text, end='\n'):
        print("\x1b[31m" + str(text) + "\x1b[0m", end=end)

    def inspect_history(self, n=1):
        last_prompt = None
        printed = []

        for x in reversed(self.history[-100:]):
            prompt = x['prompt']

            if prompt != last_prompt:
                printed.append((prompt, x['response']['choices']))
            
            last_prompt = prompt

            if len(printed) >= n:
                break
                
        for prompt, choices in reversed(printed):
            print('\n\n\n')
            print(prompt, end='')
            self.print_green(choices[0]['text'], end='')

            if len(choices) > 1:
                self.print_red(f" \t (and {len(choices)-1} other completions)", end='')
            print('\n\n\n')


    def __call__(self, prompt, only_completed=True, return_sorted=False, **kwargs):
        assert only_completed, "for now"
        assert return_sorted is False, "for now"

        if kwargs.get('n', 1) > 1:
            kwargs = {**kwargs, 'logprobs': 5}

        response = self.request(prompt, **kwargs)
        choices = response['choices']

        completed_choices = [c for c in choices if c['finish_reason'] != 'length']

        if only_completed and len(completed_choices):
            choices = completed_choices

        completions = [c['text'] for c in choices]

        if return_sorted and kwargs.get('n', 1) > 1:
            scored_completions = []

            for c in choices:
                tokens, logprobs = c['logprobs']['tokens'], c['logprobs']['token_logprobs']

                if '<|endoftext|>' in tokens:
                    index = tokens.index('<|endoftext|>') + 1
                    tokens, logprobs = tokens[:index], logprobs[:index]

                avglog = sum(logprobs) / len(logprobs)
                scored_completions.append((avglog, c['text']))

            scored_completions = sorted(scored_completions, reverse=True)
            completions = [c for _, c in scored_completions]

        return completions


@CacheMemory.cache
def cached_gpt3_request_v2(**kwargs):
    return openai.Completion.create(**kwargs)


@functools.lru_cache(maxsize=None if cache_turn_on else 0)
@NotebookCacheMemory.cache
def cached_gpt3_request_v2_wrapped(**kwargs):
    return cached_gpt3_request_v2(**kwargs)


cached_gpt3_request = cached_gpt3_request_v2_wrapped



import dsp
import copy

from functools import wraps

def shallow_copy_example_args(func):
    @wraps(func)
    def wrapper(*args, **kwargs):
        args = [dsp.Example(arg) if isinstance(arg, dsp.Example) else arg for arg in args]
        kwargs = {key: dsp.Example(value) if isinstance(value, dsp.Example) else value for key, value in kwargs.items()}
        return func(*args, **kwargs)
    return wrapper

transformation = shallow_copy_example_args



def passages2text(passages):
    if type(passages) is str:
        return passages

    assert type(passages) in [list, tuple]

    if len(passages) == 1:
        return f"«{passages[0]}»"

    return '\n'.join([f"[{idx+1}] «{txt}»" for idx, txt in enumerate(passages)])



import functools
import requests

from dsp.modules.cache_utils import CacheMemory, NotebookCacheMemory
from dsp.utils import dotdict


# TODO: Ideally, this takes the name of the index and looks up its port.


class ColBERTv2:
    def __init__(self, url='http://0.0.0.0', port=None):
        self.url = f'{url}:{port}' if port else url
        self.headers = {"Content-Type": "application/json; charset=utf-8"}
    
    def post(self, query: str, k: int):
        payload = {"query": query, "k": k}
        res = requests.post(self.url, json=payload, headers=self.headers)

        return res.json()['topk'][:k]

    def __call__(self, query: str, k=10, simplify=False):
        topk = colbertv2_get_request(self.url, query, k)
        topk = [dotdict(psg) for psg in topk]

        if simplify:
            topk = [psg.long_text for psg in topk]

        return topk


@CacheMemory.cache
def colbertv2_get_request_v2(url: str, query: str, k: int):
    assert k <= 100, f'Only k <= 100 is supported for the hosted ColBERTv2 server at the moment.'

    payload = {"query": query, "k": k}
    res = requests.get(url, params=payload)

    topk = res.json()['topk'][:k]
    topk = [{**d, 'long_text': d['text']} for d in topk]

    return topk[:k]


@functools.lru_cache(maxsize=None)
@NotebookCacheMemory.cache
def colbertv2_get_request_v2_wrapped(*args, **kwargs):
    return colbertv2_get_request_v2(*args, **kwargs)


colbertv2_get_request = colbertv2_get_request_v2_wrapped



from contextlib import contextmanager
from dsp.utils.utils import dotdict


class Settings(object):
    _instance = None

    def __new__(cls):
        """
        Singleton Pattern. See https://python-patterns.guide/gang-of-four/singleton/
        """
        
        if cls._instance is None:
            cls._instance = super().__new__(cls)
            cls._instance.stack = []

            config = dotdict()
            cls._instance.__append(config)

        return cls._instance

    @property
    def config(self):
        return self.stack[-1]

    def __getattr__(self, name):
        if hasattr(self.config, name):
            return getattr(self.config, name)
        
        if name in self.config:
            return self.config[name]

        super().__getattr__(name)

    def __append(self, config):
        self.stack.append(config)

    def __pop(self):
        self.stack.pop()
    
    def configure(self, inherit_config=True, **kwargs):
        if inherit_config:
            config = {**self.config, **kwargs}

        self.__append(config)

    @contextmanager
    def context(self, inherit_config=True, **kwargs):
        self.configure(inherit_config=inherit_config, **kwargs)

        try:
            yield
        finally:
            self.__pop()

    def __repr__(self) -> str:
        return repr(self.config)

settings = Settings()



from setuptools import setup, find_packages

setup(
    name="dsp-ml",
    version="0.1.2",
    description="Demonstrate-Search-Predict",
    url="https://github.com/stanfordnlp/dsp",
    author="Omar Khattab",
    author_email="okhattab@stanford.edu",
    license="MIT License",
    packages=find_packages(),
    python_requires='>=3.8',
    install_requires=[
        "backoff",
        "joblib",
        "jupyter",
        "openai",
        "pandas",
        "spacy",
        "regex",
    ],
    classifiers=[
		"Development Status :: 3 - Alpha",
		"Intended Audience :: Science/Research",
        "License :: OSI Approved :: MIT License",
        "Operating System :: POSIX :: Linux",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
    ],
)



import dsp
from collections import Counter
from dsp.utils import zipstar, normalize_text


class Completions:
    def __init__(self, completions: list, template):
        self.data = completions
        self.template = template

    def __iter__(self):
        return self.data.__iter__()

    def __getitem__(self, item):
        return self.data[item]

    def __len__(self):
        return len(self.data)

    def unpack(self, key=None):
        if key:
            return [getattr(c, key) for c in self.data]

        return zipstar(self.data)

    def __getattr__(self, name):
        assert len(self.data) == 1

        completion = self.data[0]

        if name in completion.keys():
            return getattr(completion, name)
        
        if name.endswith('s') and name[:-1] in completion.keys():
            pass

        assert False, name



def generate(template, **kwargs):
    generator = dsp.settings.lm

    def do_generate(example, stage, max_depth=2):
        assert stage is not None

        # Look up the appropriate fields in each demonstration.
        example = example.demos_at(lambda d: d[stage])
        
        # Generate and extract the fields.
        prompt = template(example)
        completions = generator(prompt, **kwargs)
        completions = [template.extract(example, p) for p in completions]

        # Find the completions that are most complete.
        field_names = [field.input_variable for field in template.fields]

        last_field_idx = 0
        for field_idx, key in enumerate(field_names):
            completions_ = [c for c in completions if key in c.keys() and c[key] is not None]

            # Filter out completions that are missing fields that are present in at least one completion.
            if len(completions_):
                completions = completions_
                last_field_idx = field_idx+1
        
        # If none of the completions is completed (i.e., none has the final field set).
        if last_field_idx < len(field_names):
            # Pick the first completion that has gone farthest.
            completion = completions[0]
            completion[field_names[last_field_idx]] = ''

            # Recurse with greedy decoding and a shorter length.
            max_tokens = kwargs.get('max_tokens', dsp.settings.lm.kwargs['max_tokens'])
            max_tokens = min(max(75, max_tokens // 2), max_tokens)
            new_kwargs = {**kwargs, 'max_tokens': max_tokens, 'n': 1, 'temperature': 0.0}

            assert max_depth > 0
            return generate(template, **new_kwargs)(completion, stage=stage, max_depth=max_depth-1)

        completions = Completions(completions, template=template)
        example = example.copy(completions=completions)

        if len(completions) == 1:
            completion = completions[0]
            example[stage] = example.copy(**completion)

        return example, completions
    
    return do_generate


def generate_sc(example, prompt, normalize=True, extract=None, prediction_field=None, **kwargs):
    kwargs = {'temperature': 0.7, 'max_tokens': 150, 'n': 20, **kwargs}
    completions = dsp.settings.lm(prompt, **kwargs)
    completions = extract_final_answer(example, completions, extract=extract)
    return majority_vote_(completions, normalize=normalize, prediction_field=prediction_field)

def extract_final_answer(example, completions, extract=None):
    if extract:
        completions = [extract(example, p) for p in completions]
    else:
        completions = [p.strip().split('\n')[-1].split(':', 1)[-1].strip() for p in completions]

    dsp.settings.lm.history.append({**dsp.settings.lm.history[-1], 'completions': completions})

    return completions

def majority(completions, normalize=True, field=None):
    field = completions.template.fields[-1].output_variable if field is None else field

    return Completions(majority_vote_(completions, normalize=normalize, prediction_field=field),
                      template=completions.template)


def majority_vote_(completions, normalize: bool, prediction_field: str):
    if normalize:
        original_completions = completions
        completions_ = []
        for pred in completions:
            if prediction_field in pred:
                completions_.append(normalize_text(pred[prediction_field]))
            else:
                completions_.append('')
        completions = completions_
        normalized_to_original = {}

        for completion, normalized_completion in zip(original_completions, completions):
            if normalized_completion not in normalized_to_original:
                normalized_to_original[normalized_completion] = completion

    completions_ = [x for x in completions if x]
    # completions_ = [x for x in completions_ if len(x.split()) < 7]

    if completions_:
        completions = completions_

    topk = Counter(completions).most_common()
    pred, _ = topk[0]

    if normalize:
        pred = normalized_to_original[pred]

    dsp.settings.lm.history.append({**dsp.settings.lm.history[-1], 'topk': topk, 'completions': [pred]})

    return [pred]





import dsp


def retrieve(query: str, k):
    passages = dsp.settings.rm(query, k=k)
    passages = [psg.long_text for psg in passages]

    return passages

def retrieveEnsemble(queries, k, by_prob=True):
    queries = [q for q in queries if q]

    passages = {}
    for q in queries:
        for psg in dsp.settings.rm(q, k=k*3):
            if by_prob:
                passages[psg.long_text] = passages.get(psg.long_text, 0.0) + psg.prob
            else:
                passages[psg.long_text] = passages.get(psg.long_text, 0.0) + psg.score

    passages = [(score, text) for text, score in passages.items()]
    passages = sorted(passages, reverse=True)[:k]
    passages = [text for _, text in passages]

    return passages



import tqdm
import pandas as pd

from IPython.display import display
from dsp.utils import EM

def evaluate(fn, dev, metric=EM):
    data = []

    for example in tqdm.tqdm(dev):
        question = example.question
        prediction = fn(question)

        d = dict(example)

        d['prediction'] = prediction
        d['correct'] = metric(prediction, example.answer)
        data.append(d)

    df = pd.DataFrame(data)

    print(f"Answered {df['correct'].sum()} / {len(dev)} correctly.")
    df['correct'] = df['correct'].apply(lambda x: '✔️' if x else '❌')

    pd.options.display.max_colwidth = None
    display(df.style.set_table_styles([{'selector': 'th', 'props': [('text-align', 'left')]}, {'selector': 'td', 'props': [('text-align', 'left')]}]))
